using System;
using System.Collections;

namespace RepPP.Toolkit {

	//******************************************************************
	//
	/// <summary>
	/// DeleteReference:    Handles Database integrity
	/// </summary>
	// 
	// Author:			   Jacques Fourniers, Khaled Renno
	//
	// (c) 2003-2004, Consyst SQL Inc.
	//
	//******************************************************************
	public class DeleteReference {
		/// <summary>
		/// Main table to delete
		/// </summary>
		private SCTableToDelete m_tableToDelete;
        /// <summary>
        /// Part of the where clause identifying the item to delete
        /// </summary>
        private string          m_strPKVal;
        /// <summary>
        /// Array of child tables to delete
        /// </summary>
        private ArrayList       m_arrChildTableToDelete;
        /// <summary>
        /// Array of references
        /// </summary>
        private ArrayList       m_arrReference;
        /// <summary>
        /// Rep++ Application error message
        /// </summary>
		private	string			m_strApplicationErrorMessage;

        //********************************************************************************
        //
        /// <summary>
        /// ctor:          Constructor for the class
        /// </summary>
        //
        //********************************************************************************
        public DeleteReference() {
		}

        //********************************************************************************
        //
        /// <summary>
        /// ctor:				        Constructor for the class
        /// </summary>
        /// <param name="strTableName"> Name of the table which has to be deleted</param>
        /// <param name="arrPKColName"> Array of column member of the primary key</param>
        /// <param name="arrPKVal">     Values of the primary key columns</param>
        //
        //********************************************************************************
        public DeleteReference(string strTableName, string [] arrPKColName, string[] arrPKVal) {
            m_tableToDelete         = new SCTableToDelete(strTableName, arrPKColName, arrPKVal);
            m_arrChildTableToDelete = new ArrayList();
            m_arrReference          = new ArrayList();
        }

        //********************************************************************************
        //
        /// <summary>
        /// ctor:				    Constructor for the class
        /// </summary>
        /// <param name="cinst">    Card instance to use to build the list of tables which have row deleted</param>
        //
        //********************************************************************************
        public DeleteReference(RepPP.CardInstance cinst) {
            RepPP.CardGroup cgrp;
            
            cgrp                    = cinst.Card.ChildCardGroups[0];
            m_tableToDelete         = BuildTableToDeleteFromGroup(cgrp.Group);
            m_arrChildTableToDelete = new ArrayList();
            m_arrReference          = new ArrayList();
            m_strPKVal              = "";
            BuildChildFromCardGroup(cgrp);
			//TODO: no need for this anymore as it is being handled in IsAnyReference module
            foreach (RepPP.Field fld in cgrp.Group.Fields) {
                if (m_strPKVal.Length != 0) {
                    m_strPKVal += " AND ";
                }
                m_strPKVal += "B." + fld.ColName + " = :" + fld.Group.Name + fld.Name;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// ApplicationErrorMessage     Error message
        /// </summary>
        //
        //********************************************************************************
        public string ApplicationErrorMessage {
			get {
				return(m_strApplicationErrorMessage);
			}
			set {
				m_strApplicationErrorMessage = value;
			}
		}

        //********************************************************************************
        //
        /// <summary>
        /// GetGroupInfo:		        Returns table and PK col information about a group
        /// </summary>
        /// <param name="grp">          Group for which to return the information</param>
        /// <param name="arrPKColName"> Returned the array of primary key columns</param>
        /// <returns>
        /// The table attached to the group
        /// </returns>
        //
        //********************************************************************************
        private string GetGroupInfo(RepPP.Group grp, out string[] arrPKColName) {
            string  strRetVal = "";
            int     iCount;

            iCount = 0;            
            foreach (RepPP.Field fld in grp.Fields) {
                if (fld.IsPrimaryKey) {
                    strRetVal = fld.TableName;
                    iCount++;
                }
            }
            arrPKColName = new string[iCount];
            iCount = 0;            
            foreach (RepPP.Field fld in grp.Fields) {
                if (fld.IsPrimaryKey) {
                    arrPKColName[iCount] = fld.ColName;
                    iCount++;
                }
            }
            return(strRetVal);
        }        

        //********************************************************************************
        //
        /// <summary>
        /// BuildTableToDeleteFromGroup         Build a SCTableToDelete object using a group
        /// </summary>
        /// <param name="grp">                  Group to use</param>
        /// <returns>
        /// New SCTableToDelete object
        /// </returns>
        //
        //********************************************************************************
        private SCTableToDelete BuildTableToDeleteFromGroup(RepPP.Group grp) {
            string          strTableName;
            string[]        arrPKColName;

            strTableName = GetGroupInfo(grp, out arrPKColName);
            return(new SCTableToDelete(strTableName, arrPKColName));
        }

        //********************************************************************************
        //
        /// <summary>
        /// BuildChildFromCardGroup:        Build the array of child DeleteObject using a card group hierarchy
        /// </summary>
        /// <param name="cgrp">             Card group for which to add the child</param>
        //
        //********************************************************************************
        private void BuildChildFromCardGroup(RepPP.CardGroup cgrp) {
            SCTableToDelete tableToDelete;
            
            foreach(RepPP.CardGroup cgrpChild in cgrp.ChildCardGroups) {
                tableToDelete = BuildTableToDeleteFromGroup(cgrpChild.Group);
                m_arrChildTableToDelete.Add(tableToDelete);
            }
            foreach(RepPP.CardGroup cgrpChild in cgrp.ChildCardGroups) {
                BuildChildFromCardGroup(cgrpChild);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// AddChildTableToDelete:		Add a child in the list of child table to delete
        /// </summary>
        /// <param name="strTableName"> Table name to add</param>
        /// <param name="arrPKColName"> Array of primary key for this child</param>
        //
        //********************************************************************************
        public void AddChildTableToDelete(string strTableName, string[] arrPKColName) {
            SCTableToDelete tableToDelete;

            tableToDelete = new SCTableToDelete(strTableName, arrPKColName);
            m_arrChildTableToDelete.Add(tableToDelete);
        }

        //********************************************************************************
        //
        /// <summary>
        /// AddTableReference:		            Add a reference to be checked 
        /// </summary>
        /// <param name="strTableName">         Table referencing a table to delete</param>
        /// <param name="strTableReferenced">   Table pointed to</param>
        /// <param name="strColInfoKey">        Column to return for each reference found</param>
        /// <param name="arrRefCol">            List of column pointing to the table to be deleted</param>
        //
        //********************************************************************************
        public void AddTableReference(string strTableName, string strTableReferenced, string strColInfoKey, string[] arrRefCol) {
            SCTableToDelete tableReferenced = null;
            
            if (strTableReferenced.ToUpper() == m_tableToDelete.m_strTableName.ToUpper()) {
                tableReferenced = m_tableToDelete;
            } else {
                foreach (SCTableToDelete tableToDelete in m_arrChildTableToDelete) {
                    if (strTableReferenced.ToUpper() == tableToDelete.m_strTableName.ToUpper()) {
                        tableReferenced = tableToDelete;
                        break;
                    }
                }
                if (tableReferenced == null) {
                    throw new System.ApplicationException("The table '" + strTableName + "' must be included in a table to be deleted!");
                }
            }
            m_arrReference.Add(new SCReference(strTableName, tableReferenced, strColInfoKey, arrRefCol));
        }

        //********************************************************************************
        //
        /// <summary>
        /// IsAnyReference:						Check if there is any reference to any of the child to delete
        /// </summary>
        /// <param name="connection">           Database connection</param>
        /// <returns>
        /// true if yes, false if no
        /// </returns>
        //
        //********************************************************************************
        public virtual bool IsAnyReference(RepPP.Connection connection) {
            bool				bRetVal				= false;
			bool				bIsCursorNotEmpty	= false;
            string				strSqlCmd			= null;
			RepPP.SqlCommand    sqlCommand			= null;
			RepPP.SqlCursor     sqlCursor			= null;
            int					iCount;

			ApplicationErrorMessage = "\n";
            foreach (SCReference reference in m_arrReference) {
                strSqlCmd = "SELECT A." + reference.m_strColInfoKey +
                            " FROM "  + reference.m_strTableName +
                            " A," + reference.m_tableReferenced.m_strTableName +
                            " B WHERE ";
                iCount = reference.m_arrRefCol.Length;
                for (int iIndex = 0; iIndex < iCount; iIndex++) {
                    strSqlCmd += "A." + reference.m_arrRefCol[iIndex] +
                                 " = B." + reference.m_tableReferenced.m_arrPKColName[iIndex] + " AND ";
                }
                iCount = reference.m_tableReferenced.m_arrPKColName.Length;
                for (int iIndex = 0; iIndex < iCount; iIndex++) {
                    strSqlCmd += "B."    + reference.m_tableReferenced.m_arrPKColName[iIndex] 
								 + " = " + reference.m_tableReferenced.m_arrPKVal[iIndex];
					if (iIndex != (iCount - 1)) {
						strSqlCmd += " AND ";
					}
                }
			    sqlCommand = connection.SqlCommands.Open(strSqlCmd, RepPP.SqlOptimization.sdOptFull);
			    using(sqlCommand) {
					if (sqlCommand.Execute(ref sqlCursor) != 0) {
						throw new System.Exception(connection.ErrorMessage);
					}
					using(sqlCursor) {
					    bIsCursorNotEmpty	= (sqlCursor.Next() == 0);
					    if (bIsCursorNotEmpty) {
                            ApplicationErrorMessage += "\nTable: " + reference.m_strTableName + "\n";
                            do {
                                ApplicationErrorMessage +=  "    " + sqlCursor.GetColumnValue(0) + "\n";
                            } while (sqlCursor.Next() == 0);
					    }
					    bRetVal = (bRetVal || bIsCursorNotEmpty);
					}
				}
            }
            return(bRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// DeleteThisTable:			Deletes the columns identified by the PK value
        /// </summary>
        /// <param name="connection">   Database connection</param>
        /// <param name="tableToDelete">Table to delete</param>
        //
        //********************************************************************************
        private void DeleteThisTable(RepPP.Connection connection, SCTableToDelete tableToDelete) {
            string				strSqlCmd;
			RepPP.SqlCommand    sqlCommand      = null;
			RepPP.SqlCursor     sqlCursor		= null;

            strSqlCmd = "DELETE FROM " + tableToDelete.m_strTableName + " B WHERE " + m_strPKVal;
			sqlCommand = connection.SqlCommands.Open(strSqlCmd, RepPP.SqlOptimization.sdOptFull);
			using(sqlCommand) {
			    if (sqlCommand.Execute(ref sqlCursor) != 0) {
					throw new System.Exception(connection.ErrorMessage);
				}
				if (sqlCursor != null) {
					sqlCursor.Close();
				}
			}
        }

        //********************************************************************************
        //
        /// <summary>
        /// DeleteTables:			Deletes the columns identified by the PK value
        /// </summary>
        /// <param name="cinst">    Card instance</param>
        /// <returns>
        /// true if no error, false otherwise
        /// </returns>
        //
        //********************************************************************************
        public bool DeleteTables(RepPP.CardInstance cinst) {
			bool				bRetVal;
			int					iError;
			
			if (cinst.RootGroupInstance.Type == RepPP.GroupType.sdVertical) {
				bRetVal = ((iError = cinst.DeleteFromDb()) == 0);
			} else {
				bRetVal = ((iError = cinst.UpdateToDb()) == 0);
			}
			if (!bRetVal) {
				ApplicationErrorMessage = cinst.Application.DataConnection.ErrorMessage + " " + iError;
			}
			return(bRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// DeleteTables:            Deletes the columns identified by the PK value 
        /// </summary>
        /// <param name="connection">Database connection</param>
        //
        //********************************************************************************
        public void DeleteTables(RepPP.Connection connection) {
            int     iCount;

            iCount = m_arrChildTableToDelete.Count;
            for (int iIndex = iCount - 1; iIndex >= 0; iIndex--) {
                DeleteThisTable(connection, (SCTableToDelete)m_arrChildTableToDelete[iIndex]);
            }
            DeleteThisTable(connection, m_tableToDelete);
        }

		//******************************************************************
		//
        /// <summary>
        /// SCTableToDelete:  Represent a table which content has to be deleted for the specified primary key value
        /// </summary>
		//
		//******************************************************************
        private class SCTableToDelete {
            public  string      m_strTableName;             // Table to delete
            public  string[]    m_arrPKColName;             // column part of the primary key
			public  string[]	m_arrPKVal;					// primary key values

            //********************************************************************************
            //
            /// <summary>
            /// ctor:				        Constructor for the class
            /// </summary>
            /// <param name="strTableName"> Table which content has to be deleted</param>
            /// <param name="arrPKColName"> List of column which constitute the primary key</param>
            //
            //********************************************************************************
            public SCTableToDelete(string strTableName, string[] arrPKColName) {
                m_strTableName  = strTableName;
                m_arrPKColName  = arrPKColName;
            }

            //********************************************************************************
            //
            /// <summary>
            /// ctor:				        Constructor for the class 
            /// </summary>
            /// <param name="strTableName"> Table which content has to be deleted</param>
            /// <param name="arrPKColName"> List of column which constitute the primary key</param>
            /// <param name="arrPKVal">     List of primary keys' value</param>
            //
            //********************************************************************************
            public SCTableToDelete(string strTableName, string[] arrPKColName, string[] arrPKVal) {
                m_strTableName  = strTableName;
                m_arrPKColName  = arrPKColName;
                m_arrPKVal		= arrPKVal;
            }
        };/*SCTableToDelete*/

		//******************************************************************
		//
        /// <summary>
        /// SCReference:  Represent a reference to one of the table to delete 
        /// </summary>
		//
		//******************************************************************
        private class SCReference {
            /// <summary>Table pointing to the table to be deleted</summary>
            public  string          m_strTableName;
            /// <summary>Table which has to be deleted</summary>
            public  SCTableToDelete m_tableReferenced;
            /// <summary>The column name containing the information to be returned</summary>
            public  string          m_strColInfoKey;
            /// <summary>Array of column representing the reference.</summary>
            public  string[]        m_arrRefCol;

            //********************************************************************************
            //
            /// <summary>
            /// ctor:					        Constructor for the class 
            /// </summary>
            /// <param name="strTableName">     Table pointing to the table which has to be deleted</param>
            /// <param name="tableReferenced">  Table which has to be deleted</param>
            /// <param name="strColInfoKey">    Column which is returned in case of a found reference</param>
            /// <param name="arrRefCol">        Column containing the reference. If null, the column name is the PK col of the referenced table.</param>
            //
            //********************************************************************************
            public SCReference(string			strTableName, 
							   SCTableToDelete	tableReferenced, 
							   string			strColInfoKey, 
							   string[]			arrRefCol) {
                m_strTableName      = strTableName;
                m_tableReferenced   = tableReferenced;
                m_strColInfoKey     = strColInfoKey;
                m_arrRefCol         = (arrRefCol != null) ? arrRefCol : tableReferenced.m_arrPKColName;
                if (m_arrRefCol.Length != tableReferenced.m_arrPKColName.Length) {
                    throw new System.ApplicationException("The number of pointing column must be the same as the number of primary key column in the referenced table");
                }
            }
        };/*SCReference*/

    };/*SCDeleteReference*/
}/*namespace*/